(* camlp4r ./def.syn.cmo ./pa_html.cmo *)
(* $Id: ascend.ml,v 3.58 2001/01/31 12:34:15 ddr Exp $ *)
(* Copyright (c) 2001 INRIA *)

open Config;
open Def;
open Gutil;
open Util;

value limit_by_list conf =
  match p_getint conf.base_env "max_anc_level" with
  [ Some x -> max 1 x
  | None -> 8 ]
;

value limit_by_tree conf =
  match p_getint conf.base_env "max_anc_tree" with
  [ Some x -> max 1 x
  | None -> 7 ]
;

value niveau_max_ascendance base ip =
  let _ = base.data.ascends.array () in
  let _ = base.data.couples.array () in
  let x = ref 0 in
  let mark = Array.create base.data.persons.len False in
  do let rec loop niveau ip =
       if mark.(Adef.int_of_iper ip) then ()
       else
         do mark.(Adef.int_of_iper ip) := True;
            x.val := max x.val niveau;
         return
         match (aoi base ip).parents with
         [ Some ifam ->
             let cpl = coi base ifam in
             do loop (succ niveau) cpl.father;
                loop (succ niveau) cpl.mother;
             return ()
         | _ -> () ]
     in
     loop 0 ip;
  return x.val
;

value text_to conf =
  fun
  [ 1 ->
      transl conf "specify" ^ " " ^ transl_nth conf "generation/generations" 0
  | 2 -> transl conf "to the parents"
  | 3 -> transl conf "to the grandparents"
  | 4 -> transl conf "to the great-grandparents"
  | i ->
      Printf.sprintf (ftransl conf "to the %s generation")
        (transl_nth conf "nth (generation)" i) ]
;

value text_level conf =
  fun
  [ 1 ->
      transl conf "specify" ^ " " ^ transl_nth conf "generation/generations" 0
  | 2 -> transl conf "the parents"
  | 3 -> transl conf "the grandparents"
  | 4 -> transl conf "the great-grandparents"
  | i ->
      Printf.sprintf (ftransl conf "the %s generation")
        (transl_nth conf "nth (generation)" i) ]
;

value print_choice conf base p niveau_effectif =
  tag "form" "method=get action=\"%s\"" conf.command begin
    Util.hidden_env conf;
    Wserver.wprint "\n";
    Wserver.wprint "<input type=hidden name=m value=A>\n";
    wprint_hidden_person conf base "" p;
    tag "center" begin
      tag "select" "name=v" begin
        let rec boucle i =
          if i > niveau_effectif + 1 then ()
          else
            do Wserver.wprint "  <option value=%d%s> %s\n" i
                 (if i == 0 then " selected" else "")
                 (capitale (text_to conf i));
            return boucle (succ i)
        in
        boucle 1;
      end;
      Wserver.wprint "<input type=submit value=\"Ok\">\n";
    end;
    html_br conf;
    tag "table" "border=%d width=\"90%%\"" conf.border begin
      tag "tr" begin
        tag "td" "valign=top" begin
          Wserver.wprint "<input type=radio name=t value=N checked> %s<br>\n"
            (capitale (transl conf "short display"));
          Wserver.wprint "<input type=radio name=t value=G> %s<br>\n"
            (capitale (transl conf "long display"));
          Wserver.wprint
            "- %s <input type=checkbox name=siblings value=on checked><br>\n"
            (capitale (transl conf "siblings"));
          Wserver.wprint
            "- %s <input type=checkbox name=notes value=on checked><br>\n"
            (capitale (nominative (transl_nth conf "note/notes" 1)));
          Wserver.wprint "<input type=radio name=t value=T> %s\n"
            (capitale (transl conf "tree"));
          let limit =
            (* if browser_doesnt_have_tables conf then 3 else *)
            limit_by_tree conf
          in
          if niveau_effectif <= limit then ()
          else
            Wserver.wprint "(%s %d %s)\n" (transl conf "maximum") limit
              (transl_nth conf "generation/generations" 1);
          Wserver.wprint "<br>\n";
          Wserver.wprint "<input type=radio name=t value=A> %s\n"
            (capitale (transl_nth conf "male line/female line" 0));
          Wserver.wprint "<br>\n";
          Wserver.wprint "<input type=radio name=t value=C> %s\n"
            (capitale (transl_nth conf "male line/female line" 1));
          Wserver.wprint "<br>\n";
          Wserver.wprint "- %s <input type=checkbox name=image value=on><br>\n"
            (capitale (transl_nth conf "image/images" 1));
        end;
        tag "td valign=top" begin
          Wserver.wprint "<input type=radio name=t value=L> %s\n"
            (capitale (transl_nth conf "list/list (ancestors)" 1));
          if niveau_effectif <= limit_by_list conf then ()
          else
            Wserver.wprint "(%s %d %s)\n" (transl conf "maximum")
              (limit_by_list conf)
              (transl_nth conf "generation/generations" 1);
          Wserver.wprint "<br>\n";
          Wserver.wprint "<input type=radio name=t value=H> %s\n"
            (capitale (transl conf "horizontally"));
          if niveau_effectif <= limit_by_list conf then ()
          else
            Wserver.wprint "(%s %d %s)\n" (transl conf "maximum")
              (limit_by_list conf)
              (transl_nth conf "generation/generations" 1);
          Wserver.wprint "<br>\n";
          Wserver.wprint "<input type=radio name=t value=S> %s<br>\n"
            (capitale (transl conf "only the generation selected"));
          Wserver.wprint "<input type=radio name=t value=F> %s\n"
            (capitale (transl conf "surnames list"));
          Wserver.wprint "<br>\n";
          Wserver.wprint "<input type=radio name=t value=M> %s<br>\n"
            (capitale (transl conf "missing ancestors"));
          Wserver.wprint "- %s\n"
            (capitale (transl conf "alphabetic order"));
          Wserver.wprint "<input type=checkbox name=al value=on><br>\n";
          Wserver.wprint "- %s\n"
            (capitale (transl conf "include missing spouses"));
          Wserver.wprint "<input type=checkbox name=ms value=on><br>\n";
          Wserver.wprint "- %s\n" (capitale (transl conf "after"));
          Wserver.wprint "<input name=after size=5 maxlength=5>\n";
          Wserver.wprint "%s\n" (capitale (transl conf "before"));
          Wserver.wprint "<input name=before size=5 maxlength=5><br>\n";
        end;
      end;
      tag "tr" begin
        tag "td" "colspan=2 align=center" begin
          Wserver.wprint "%s\n"
            (capitale (transl conf "cancel GeneWeb links"));
          Wserver.wprint "<input type=checkbox name=cgl value=on>\n";
        end;
      end;
    end;
  end
;

value afficher_menu_ascendants conf base p =
  let niveau_effectif = niveau_max_ascendance base p.cle_index in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (capitale (transl conf "ancestors"))
            (txt_fun raw_access conf base p)))
  in
  do header conf title;
     tag "center" begin
       print_choice conf base p niveau_effectif;
       html_p conf;
       Wserver.wprint
         (fcapitale (ftransl conf "navigation with %t as Sosa reference"))
         (fun _ ->
            do conf.henv := List.remove_assoc "iz" conf.henv;
               stag "a" "href=\"%siz=%d\"" (commd conf)
                 (Adef.int_of_iper p.cle_index)
               begin
                 afficher_personne_sans_titre conf base p;
               end;
               afficher_titre conf base p;
            return ());
       Wserver.wprint ".\n";
     end;
     Wserver.wprint "<br>\n";
     trailer conf;
  return ()
;

value afficher_ancetre conf base p =
  do afficher_personne_referencee conf base p;
     Date.afficher_dates_courtes conf base p;
  return ()
;

value afficher_ascendants_jusqu_a conf base niveau_max p =
  let niveau_max = min (limit_by_list conf)  niveau_max in
  let rec boucle niveau ip =
    if niveau < niveau_max then
      let x = aoi base ip in
      match x.parents with
      [ Some ifam ->
          let cpl = coi base ifam in
          let pere = poi base cpl.father in
          let mere = poi base cpl.mother in
          let know_fath = connais base pere in
          let know_moth = connais base mere in
          if know_fath || know_moth then
            tag "ul" begin
              if know_fath then
                do Wserver.wprint "<li type=square> ";
                   afficher_ancetre conf base pere;
                   Wserver.wprint "\n";
                return boucle (succ niveau) cpl.father
              else ();
              if know_moth then
                do Wserver.wprint "<li type=circle> ";
                   afficher_ancetre conf base mere;
                   Wserver.wprint "\n";
                return boucle (succ niveau) cpl.mother
              else ();
            end
          else ()
      | None -> () ]
    else ()
  in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
            (txt_fun raw_access conf base p)))
  in
  do header conf title;
     tag "nobr" begin
       Wserver.wprint "%s.\n" (capitale (text_to conf niveau_max));
       boucle 1 p.cle_index;
     end;
     trailer conf;
  return ()
;

(* Print ancestors with numbers.
   The mark table holds the number of the ancestor after it has been
   printed or Num.zero if it has not yet been printed.
   At each generation, count and print a list of generation_person *)

type generation_person =
  [ GP_person of Num.t and iper and option ifam
  | GP_same of Num.t and Num.t and iper
  | GP_interv of option (Num.t * Num.t * option (Num.t * Num.t))
  | GP_missing of Num.t and iper ]
;

value next_generation base mark gpl =
  let gpl =
    List.fold_right
      (fun gp gpl ->
         match gp with
         [ GP_person n ip _ ->
             let n_fath = Num.twice n in
             let n_moth = Num.inc n_fath 1 in
             let a = aoi base ip in
             match a.parents with
             [ Some ifam ->
                 let cpl = coi base ifam in
                 [GP_person n_fath cpl.father (Some ifam);
                  GP_person n_moth cpl.mother (Some ifam) ::
                  gpl]
             | None -> [GP_missing n ip :: gpl] ]
         | GP_interv None -> [gp :: gpl]
         | GP_interv (Some (n1, n2, x)) ->
             let x =
               match x with
               [ Some (m1, m2) -> Some (Num.twice m1, Num.twice m2)
               | None -> None ]
             in
             let gp = GP_interv (Some (Num.twice n1, Num.twice n2, x)) in
             [gp :: gpl]
         | _ -> gpl ])
      gpl []
  in
  let gpl =
    List.fold_left
      (fun gpl gp ->
         match gp with
         [ GP_person n ip _ ->
             let i = Adef.int_of_iper ip in
             let m = mark.(i) in
             if Num.eq m Num.zero then do mark.(i) := n; return [gp :: gpl]
             else [GP_same n m ip :: gpl]
         | _ -> [gp :: gpl] ])
      [] gpl
  in
  List.rev gpl
;

value wpr s = Wserver.wprint "%s" s;

value print_generation_person conf base cnt gp =
  match gp with
  [ GP_person n ip _ ->
      let p = poi base ip in
      do html_li conf;
         Num.print wpr (transl conf "(thousand separator)") n;
         Wserver.wprint " -\n";
         afficher_personne_titre_referencee conf base p;
         Date.afficher_dates_courtes conf base p;
         Wserver.wprint "\n";
         incr cnt;
      return ()
  | GP_same n1 n2 ip ->
      let p = poi base ip in
      do html_li conf;
         Num.print wpr (transl conf "(thousand separator)") n1;
         Wserver.wprint " =&gt; ";
         let s =
           let b = ref "" in
           do Num.print (fun s -> b.val := b.val ^ s)
                (transl conf "(thousand separator)") n2;
           return b.val
         in
         Wserver.wprint "%s" (reference conf base p s);
         Wserver.wprint "\n\n";
      return ()
  | GP_interv None ->
      do html_li conf;
         Wserver.wprint "...\n\n";
      return ()
  | GP_interv (Some (n1, n2, x)) ->
      do html_li conf;
         Num.print wpr (transl conf "(thousand separator)") n1;
         Wserver.wprint "-";
         Num.print wpr (transl conf "(thousand separator)")
           (Num.sub n2 Num.one);
         Wserver.wprint " = ";
         match x with
         [ Some (m1, m2) ->
             do Num.print wpr (transl conf "(thousand separator)") m1;
                Wserver.wprint "-";
                Num.print wpr (transl conf "(thousand separator)")
                  (Num.sub m2 Num.one);
             return ()
         | None -> Wserver.wprint "..." ];
         Wserver.wprint "\n\n";
      return ()
  | _ -> () ]
;

value will_print =
  fun
  [ GP_person _ _ _ -> True
  | GP_same _ _ _ -> True
  | _ -> False ]
;

value afficher_ascendants_numerotation conf base niveau_max p =
  let mark = Array.create (base.data.persons.len) Num.zero in
  let cnt = ref 0 in
  let rec generation niveau gpl =
    if niveau <= niveau_max then
      do html_li conf;
         Wserver.wprint "%s %d\n"
           (capitale (transl_nth conf "generation/generations" 0)) niveau;
         tag "ul" begin
           List.iter (print_generation_person conf base cnt) gpl;
         end;
      return
      let gpl = next_generation base mark gpl in
      if List.exists will_print gpl then generation (niveau + 1) gpl else ()
    else ()
  in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
            (txt_fun raw_access conf base p)))
  in
  do header conf title;
     print_link_to_welcome conf True;
     Wserver.wprint "%s.\n" (capitale (text_to conf niveau_max));
     tag "ul" begin
       mark.(Adef.int_of_iper p.cle_index) := Num.one;
       generation 1 [GP_person Num.one p.cle_index None];
     end;
     if cnt.val > 1 then
       do html_p conf;
          Wserver.wprint "%s: %d %s\n" (capitale (transl conf "total")) cnt.val
            (transl_nth conf "person/persons" 1);
       return ()
     else ();
     trailer conf;
  return ()
;

value link_long_txt conf n =
  let s =
    let r = ref "" in
    let print s = r.val := r.val ^ s in
    do Num.print print (transl conf "(thousand separator)") n; return
    r.val
  in
  "<strong><a href=\"#" ^ Num.to_string n ^ "\">" ^ s ^ "</a></strong>"
;

value print_link_long conf n =
  Wserver.wprint "%s" (link_long_txt conf n)
;

value print_person_long_info conf base auth link p =
  do List.iter
       (fun a ->
          Wserver.wprint ", %s <em>%s</em>" (nominative (transl conf "alias"))
            (sou base a))
       p.aliases;
     if link = None && List.length p.titles > 0 &&
       (p.access <> Private || conf.friend || conf.wizard)
     then
       do Wserver.wprint ", <em>";
          Perso.print_titles conf base False (transl conf "and") p;
          Wserver.wprint "</em>";
       return ()
     else ();
     if auth then Date.print_dates conf base p else ();
     let occu = sou base p.occupation in
     if auth && link = None && occu <> "" then
       Wserver.wprint ", %s" occu
     else ();
     match link with
     [ Some n ->
         do Wserver.wprint ".\n %s " (capitale (transl conf "see"));
            print_link_long conf n;
         return ()
     | None -> () ];
  return ()
;

value title_reference conf base t =
  let ident = sou base t.t_ident in
  let place = sou base t.t_place in
  let s = if place = "" then ident else ident ^ " " ^ place in
  "<em>" ^
  geneweb_link conf
    ("m=TT;sm=S;t=" ^ code_varenv ident ^ ";p=" ^ code_varenv place)
    s ^
  "</em>"
;

value strong_referenced_person_title_text conf base p =
  if p.access <> Private || conf.friend || conf.wizard then
    match main_title base p with
    [ Some t ->
        "<strong>" ^ reference conf base p (titled_person_text conf base p t) ^
        "</strong>,\n" ^ title_reference conf base t
    | None ->
        "<strong>" ^ reference conf base p (person_text conf base p) ^
        "</strong>" ]
  else
    "<strong>" ^ reference conf base p (person_text conf base p) ^ "</strong>"
;

value get_link all_gp ip =
  loop all_gp where rec loop =
    fun
    [ [GP_person n ip0 _ :: gpl] ->
        if ip = ip0 then Some n else loop gpl
    | [gp :: gpl] -> loop gpl
    | [] -> None ]
;

value print_persons_parents conf base all_gp p =
  let a = aoi base p.cle_index in
  match a.parents with
  [ Some ifam ->
      let cpl = coi base ifam in
      let string ip =
        let p = poi base ip in        
        let link = get_link all_gp ip in
        strong_referenced_person_title_text conf base p ^
        Date.short_dates_text conf base p ^
        (match link with
         [ Some n ->
             "\n(" ^ transl conf "see" ^ " " ^ link_long_txt conf n ^ ")"
         | None -> "" ])
      in
      Wserver.wprint ", %s\n"
        (transl_decline2 conf
           "%1 of (same or greater generation level) %2"
           (transl_nth conf "son/daughter/child" (index_of_sex p.sex))
           (string cpl.father ^ "\n" ^ transl conf "and" ^ "\n" ^
            string cpl.mother))
  | None -> () ]
;

value print_marriage_long conf base all_gp auth marr_nb p ifam =
  let fam = foi base ifam in
  let ispouse = spouse p.cle_index (coi base ifam) in
  let spouse = poi base ispouse in
  let divorce = fam.divorce in
  let auth = auth && age_autorise conf base spouse in
  do Wserver.wprint (fcapitale (relation_txt conf p.sex fam))
       (fun _ ->
          do match marr_nb with
             [ Some n -> Wserver.wprint " (%d)" n
             | None -> () ];
             if auth then Perso.print_marriage_text conf base False fam
             else ();
          return ());
     Wserver.wprint "\n";
     stag "strong" begin
       Wserver.wprint "%s"
         (reference conf base spouse (person_text conf base spouse));
     end;
     print_person_long_info conf base auth (get_link all_gp ispouse) spouse;
     print_persons_parents conf base all_gp spouse;
     Wserver.wprint ".\n";
     match divorce with
     [ Divorced d ->
         let d = Adef.od_of_codate d in
         do Wserver.wprint "\n%s" (capitale (transl conf "divorced"));
            match d with
            [ Some d when auth ->
                Wserver.wprint " %s" (Date.string_of_ondate conf d)
            | _ -> () ];
            Wserver.wprint ".\n";
         return ()
     | _ -> () ];
  return ()
;

value person_has_notes base p =
  sou base p.notes <> "" || sou base p.psources <> "" ||
  sou base p.birth_src <> "" || sou base p.baptism_src <> "" ||
  sou base p.death_src <> "" || sou base p.burial_src <> "" ||
  let u = uoi base p.cle_index in
  List.exists
    (fun ifam ->
       let fam = foi base ifam in
       sou base fam.marriage_src <> "" || sou base fam.marriage_src <> "")
    (Array.to_list u.family)
;
 
value has_notes conf base =
  List.exists
    (fun
     [ GP_person _ ip _ ->
         let p = poi base ip in
         age_autorise conf base p && person_has_notes base p
     | _ -> False ])
;

value print_other_marriages conf base ws all_gp auth ifamo p u =
  if ws && (ifamo = None || Array.length u.family >= 2) then
    do for i = 0 to Array.length u.family - 1 do
         if ifamo = None || ifamo <> Some u.family.(i) then
           let marr_nb = if ifamo = None then None else Some (i + 1) in
           print_marriage_long conf base all_gp auth marr_nb p u.family.(i)
         else ();
       done;
    return ()
  else ()
;

value print_notes_ref conf base p wn n child_n =
  if wn && age_autorise conf base p && person_has_notes base p then
    do Wserver.wprint "[%s "
         (capitale (transl_nth conf "note/notes" 0));
       stag "strong" begin
         stag "a" "href=\"#notes-%s%s\"" (Num.to_string n) child_n begin
           Num.print wpr (transl conf "(thousand separator)") n;
           Wserver.wprint "%s" child_n;
         end;
       end;
       Wserver.wprint "].\n";
    return ()
  else ()
;

value print_family_long conf base ws wn all_gp ifam nth moth_nb =
  let cpl = coi base ifam in
  let des = doi base ifam in
  let auth =
    age_autorise conf base (poi base cpl.father) &&
    age_autorise conf base (poi base cpl.mother)
  in
  do Wserver.wprint "<p>\n... ";
     Wserver.wprint "%s" (transl conf "having as children");
     match nth with
     [ Some (n, i) ->
         do Wserver.wprint " ";
            print_link_long conf n;
            Wserver.wprint "-(X%d)" i;
         return ()
     | None -> () ];
     Wserver.wprint ":\n<p>\n";
     let auth =
       List.for_all
         (fun ip -> age_autorise conf base (poi base ip))
         (Array.to_list des.children)
     in
     tag "ol" "type=a" begin
       for i = 0 to Array.length des.children - 1 do
         let ipc = des.children.(i) in
         if ws || get_link all_gp ipc <> None then
           let pc = poi base ipc in
           let uc = uoi base ipc in
           let n = get_link all_gp ipc in
           do Wserver.wprint "<li>\n";
              stag "strong" begin
                if pc.surname = (poi base cpl.father).surname then
                  afficher_prenom_de_personne_referencee conf base
                    pc
                else
                  afficher_personne_referencee conf base pc;
              end;
              print_person_long_info conf base auth n pc;
              Wserver.wprint ".\n";
              match n with
              [ Some _ -> ()
              | None ->
                  do for i = 0 to Array.length uc.family - 1 do
                       print_marriage_long conf base all_gp auth None
                         pc (uc.family.(i));
                     done;
                     let child_n =
                       let s =
                         String.make 1 (Char.chr (Char.code 'a' + i))
                       in
                       match nth with
                       [ Some (_, i) -> "-" ^ string_of_int i ^ s
                       | None -> s ]
                     in
                     print_notes_ref conf base pc wn moth_nb child_n;
                  return () ];
              if i <> Array.length des.children - 1 then
                Wserver.wprint "<br>&nbsp;\n"
              else ();
           return ()
         else ();
       done;
     end;
  return ()
;

value print_other_families conf base wn all_gp excl_ifam moth_nb =
  let print ip n =
    let u = uoi base ip in
    for i = 0 to Array.length u.family - 1 do
      let ifam = u.family.(i) in
      if ifam <> excl_ifam && Array.length (doi base ifam).children <> 0 then
        print_family_long conf base True wn all_gp ifam (Some (n, i + 1))
          moth_nb
      else ();
    done
  in
  let cpl = coi base excl_ifam in
  do print cpl.father (Num.sub moth_nb (Num.one));
     print cpl.mother moth_nb;
  return ()
;

value print_generation_person_long conf base ws wn all_gp last_gen gp =
  match gp with
  [ GP_person n ip ifamo ->
      let p = poi base ip in
      let u = uoi base ip in
      do Wserver.wprint "<p>\n";
         match ifamo with
         [ Some ifam ->
             if not (Num.even n) then
               let fam = foi base ifam in
               let cpl = coi base ifam in
               let auth =
                 age_autorise conf base (poi base cpl.father) &&
                 age_autorise conf base (poi base cpl.mother)
               in
               do Wserver.wprint "... ";
                  Wserver.wprint (relation_txt conf Male fam)
                    (fun _ ->
                       if auth then
                         Perso.print_marriage_text conf base False fam
                       else ());
                  Wserver.wprint "...\n<p>\n";
               return ()
             else ()
         | _ -> () ];
         stag "strong" begin
           stag "a" "name=\"%s\"" (Num.to_string n) begin
             Num.print wpr (transl conf "(thousand separator)") n;
           end;
         end;
         Wserver.wprint ":\n";
         stag "strong" begin
           afficher_personne_referencee conf base p;
         end;
         print_person_long_info conf base (age_autorise conf base p) None p;
         Wserver.wprint ".\n";
         match (aoi base ip).parents with
         [ Some ifam ->
             let cpl = coi base ifam in
             let prec =
               if last_gen then
                 (get_link all_gp cpl.father, get_link all_gp cpl.mother)
               else
                 let n1 = Num.twice n in
                 (Some n1, Some (Num.inc n1 1))
             in
             match prec with
             [ (Some n1, Some n2) ->
                 do Wserver.wprint "%s: "
                      (capitale (nominative (transl conf "parents")));
                    print_link_long conf n1;
                    Wserver.wprint " %s " (transl conf "and");
                    print_link_long conf n2;
                    Wserver.wprint ".\n";
                 return ()
             | _ -> () ]
         | None -> () ];
         print_notes_ref conf base p wn n "";
         print_other_marriages conf base ws all_gp (age_autorise conf base p)
           ifamo p u;
         match ifamo with
         [ Some ifam ->
             if not (Num.even n) then
               do print_family_long conf base ws wn all_gp ifam None n;
                  if ws then print_other_families conf base wn all_gp ifam n
                  else ();
               return ()
             else ()
         | _ -> () ];
      return ()
  | GP_same n1 n2 ip ->
      let p = poi base ip in
      do Wserver.wprint "<p>\n";
         stag "strong" begin
           stag "a" "name=\"%s\"" (Num.to_string n1) begin
             Num.print wpr (transl conf "(thousand separator)") n1;
           end;
         end;
         Wserver.wprint ": %s " (transl conf "see");
         print_link_long conf n2;
         Wserver.wprint ".\n\n";
      return ()
  | _ -> () ]
;

value print_not_empty_src conf base new_parag first txt isrc =
  let src = sou base isrc in
  if src = "" then ()
  else
    do if first.val then
         do if new_parag then html_p conf else ();
            Wserver.wprint "<font size=-1><em>%s:</em></font>\n"
              (capitale (transl_nth conf "source/sources" 1));
         return ()
       else ();
       html_br conf;
       Wserver.wprint "-\n";
       first.val := False;
       Wserver.wprint "<font size=-1><em>%s: " (txt ());
       copy_string_with_macros conf src;
       Wserver.wprint "</em></font>\n";
    return ()
;

value print_sources conf base new_parag p =
  let u = uoi base p.cle_index in
  let first = ref True in
  let auth = age_autorise conf base p in
  do print_not_empty_src conf base new_parag first
       (fun () -> nominative (transl_nth conf "person/persons" 0))
       p.psources;
     if auth then
       do print_not_empty_src conf base new_parag first
            (fun () -> transl_nth conf "birth" 0)
             p.birth_src;
          print_not_empty_src conf base new_parag first
            (fun () -> transl_nth conf "baptism" 0)
            p.baptism_src;
          print_not_empty_src conf base new_parag first
            (fun () -> transl_nth conf "death" 0)
            p.death_src;
          print_not_empty_src conf base new_parag first
            (fun () -> transl_nth conf "burial" 0)
            p.burial_src;
       return ()
     else ();
     for i = 0 to Array.length u.family - 1 do
       let fam = foi base u.family.(i) in
       do if auth then
            print_not_empty_src conf base new_parag first
              (fun () ->
                 (nominative (transl_nth conf "marriage/marriages" 0)) ^
                 (if Array.length u.family == 1 then ""
                  else " " ^ string_of_int (i + 1)))
               fam.marriage_src
          else ();
          print_not_empty_src conf base new_parag first
            (fun () ->
               nominative (transl_nth conf "family/families" 0) ^
               (if Array.length u.family == 1 then ""
                else " " ^ string_of_int (i + 1)))
            fam.fsources;
       return ();
     done;
  return ()
;

value print_notes_for_someone conf base p n child_n =
  if age_autorise conf base p && person_has_notes base p then
    let notes = sou base p.notes in
    do Wserver.wprint "<dt>\n";
       stag "strong" begin
         stag "a" "name=\"notes-%s%s\"" (Num.to_string n) child_n begin
           stag "a" "href=#%s" (Num.to_string n) begin
             Num.print wpr (transl conf "(thousand separator)") n;
             Wserver.wprint "%s" child_n;
           end;
         end;
       end;
       Wserver.wprint ": \n<dd>\n";
       copy_string_with_macros conf notes;
       print_sources conf base (notes <> "") p;
       Wserver.wprint "<p>\n";
    return ()
  else ()
;

value print_notes_for_family conf base all_gp ifam nth moth_nb =
  let des = doi base ifam in
  for i = 0 to Array.length des.children - 1 do
    let ipc = des.children.(i) in
    let pc = poi base ipc in
    let n = get_link all_gp ipc in
    match n with
    [ Some _ -> ()
    | None ->
        let child_n =
          let s = String.make 1 (Char.chr (Char.code 'a' + i)) in
          match nth with
          [ Some (_, i) -> "-" ^ string_of_int i ^ s
          | None -> s ]
        in
        print_notes_for_someone conf base pc moth_nb child_n ];
  done
;

value print_notes_for_other_families conf base all_gp excl_ifam moth_nb =
  let print ip n =
    let u = uoi base ip in
    for i = 0 to Array.length u.family - 1 do
      let ifam = u.family.(i) in
      if ifam <> excl_ifam && Array.length (doi base ifam).children <> 0 then
        print_notes_for_family conf base all_gp ifam (Some (n, i + 1))
          moth_nb
      else ();
    done
  in
  let cpl = coi base excl_ifam in
  do print cpl.father (Num.sub moth_nb (Num.one));
     print cpl.mother moth_nb;
  return ()
;

value print_notes conf base all_gp ws =
  fun
  [ GP_person n ip ifamo ->
      do print_notes_for_someone conf base (poi base ip) n "";
         match ifamo with
         [ Some ifam ->
             if not (Num.even n) && ws then
               do print_notes_for_family conf base all_gp ifam None n;
                  print_notes_for_other_families conf base all_gp ifam n;
               return ()
             else ()
         | _ -> () ];
      return ()
  | _ -> () ]
;

value afficher_ascendants_numerotation_long conf base niveau_max ws wn p =
  let mark = Array.create (base.data.persons.len) Num.zero in
  let rec get_generations niveau gpll gpl =
    let gpll = [gpl :: gpll] in
    if niveau < niveau_max then
      let next_gpl = next_generation base mark gpl in
      if List.exists will_print next_gpl then
        get_generations (niveau + 1) gpll next_gpl
      else gpll
    else gpll
  in
  let rec generation niveau all_gp =
    fun
    [ [gpl :: gpll] ->
        do tag "h3" begin
             Wserver.wprint "<em>%s %d</em>\n"
               (capitale (transl_nth conf "generation/generations" 0)) niveau;
           end;
           List.iter
             (print_generation_person_long conf base ws wn all_gp (gpll = []))
             gpl;
           Wserver.wprint "<p>";
           html_br conf;
        return
        generation (niveau + 1) all_gp gpll
    | [] -> () ]
  in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
            (txt_fun raw_access conf base p)))
  in
  do header conf title;
     Wserver.wprint "%s.\n" (capitale (text_to conf niveau_max));
     mark.(Adef.int_of_iper p.cle_index) := Num.one;
     let gpll = get_generations 1 [] [GP_person Num.one p.cle_index None] in
     let gpll = List.rev gpll in
     let all_gp = List.flatten gpll in
     do generation 1 all_gp gpll;
        if wn && has_notes conf base all_gp then
          do Wserver.wprint "<p><hr><p>\n";
             Wserver.wprint "<h3>%s</h3>\n"
               (capitale (nominative (transl_nth conf "note/notes" 1)));
             tag "dl" begin
               List.iter (print_notes conf base all_gp ws) all_gp;
             end;
          return ()
        else ();
     return ();
     trailer conf;
  return ()
;

value print_ancestors_same_time_descendants conf base p a =
  let maxlen =
    match p_getint conf.env "l" with
    [ Some len -> len
    | None -> -1 ]
  in
  let predic =
    let tab = Array.create base.data.persons.len False in
    let rec mark_descendants len p =
      let i = Adef.int_of_iper p.cle_index in
      if maxlen > 0 && len > maxlen then ()
      else if tab.(i) then ()
      else
        do tab.(i) := True; return
        let u = uoi base p.cle_index in
        for i = 0 to Array.length u.family - 1 do
          let des = doi base u.family.(i) in
          for i = 0 to Array.length des.children - 1 do
            mark_descendants (len + 1) (poi base des.children.(i));
          done;
        done
    in
    do mark_descendants 0 a; return
    fun ip -> tab.(Adef.int_of_iper ip)
  in
  let will_print =
    fun
    [ GP_person _ ip _ -> predic ip
    | GP_same _ _ _ -> False
    | _ -> False ]
  in
  let mark = Array.create (base.data.persons.len) Num.zero in
  let cnt = ref 0 in
  let rec generation niveau gpl =
    if List.exists will_print gpl then
      do html_li conf;
         Wserver.wprint "%s %d\n"
           (capitale (transl_nth conf "generation/generations" 0)) niveau;
         tag "ul" begin
           List.iter
              (fun gp ->
                 if will_print gp then print_generation_person conf base cnt gp
                 else ())
               gpl;
         end;
      return
      let gpl = next_generation base mark gpl in
      generation (niveau + 1) gpl
    else ()
  in
  let title h =
    if h then
      Wserver.wprint "%s... %s"
        (capitale (transl conf "ancestors"))
        (transl_decline conf "up to" "...")
    else
      do Wserver.wprint "%s\n"
           (capitale
              (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
                 (gen_person_text raw_access conf base p)));
         Wserver.wprint "%s"
           (transl_decline conf "up to" (person_text conf base a));
      return ()
  in
  do header conf title;
     conf.senv := [];
     tag "ul" begin
       mark.(Adef.int_of_iper p.cle_index) := Num.one;
       generation 1 [GP_person Num.one p.cle_index None];
     end;
     trailer conf;
  return ()
;

value afficher_ascendants_niveau conf base niveau_max p =
  let mark = Array.create (base.data.persons.len) Num.zero in
  let cnt = ref 0 in
  let rec generation niveau gpl =
    do for i = 0 to base.data.persons.len - 1 do
         mark.(i) := Num.zero;
       done;
    return
    if niveau < niveau_max then
      let gpl =
        List.map
          (fun gp ->
             match gp with
             [ GP_same n m ip ->
                 GP_interv (Some (n, Num.inc n 1, Some (m, Num.inc m 1)))
             | _ -> gp ])
          gpl
      in
      let gpl = next_generation base mark gpl in
      let gpl =
        List.fold_right
          (fun gp gpl ->
             match (gp, gpl) with
             [ (GP_interv (Some (n1, n2, x)),
                [GP_interv (Some (n3, n4, y)) :: gpl1]) ->
                 if Num.eq n2 n3 then
                   let z =
                     match (x, y) with
                     [ (Some (m1, m2), Some (m3, m4)) ->
                         if Num.eq m2 m3 then Some (m1, m4)
                         else None
                     | _ -> None ]
                   in
                   [GP_interv (Some (n1, n4, z)) :: gpl1]
                 else [GP_interv None :: gpl1]
             | (GP_interv _, [GP_interv _ :: gpl]) -> [GP_interv None :: gpl]
             | (GP_missing _ _, gpl) -> gpl
             | _ -> [gp :: gpl] ])
          gpl []
      in
      generation (niveau + 1) gpl
    else
      do html_li conf;
         Wserver.wprint "%s\n" (capitale (text_level conf niveau_max));
         tag "ul" begin
           List.iter (print_generation_person conf base cnt) gpl;
         end;
      return ()
  in
  let title h =
    if h then
      Wserver.wprint "%s %d\n"
        (capitale (transl_nth conf "generation/generations" 0)) niveau_max
    else
      Wserver.wprint "%s"
        (capitale
           (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
              (gen_person_text raw_access conf base p)))
  in
  do header conf title;
     tag "ul" begin
       mark.(Adef.int_of_iper p.cle_index) := Num.one;
       generation 1 [GP_person Num.one p.cle_index None];
     end;
     if cnt.val > 1 then
       do html_p conf;
          Wserver.wprint "%s: %d %s\n" (capitale (transl conf "total")) cnt.val
            (transl_nth conf "person/persons" 1);
       return ()
     else ();
     trailer conf;
  return ()
;

value print_generation_missing_persons conf base title sp_incl gp =
  let print_title () =        
    match title.val with
    [ Some level ->
       do html_li conf;
          Wserver.wprint "%s %d\n"
            (capitale (transl_nth conf "generation/generations" 0)) level;
          Wserver.wprint "<ul>\n";
          title.val := None;
       return ()
    | _ -> () ]
  in
  match gp with
  [ GP_person n ip _ ->
      let p = poi base ip in
      let u = uoi base ip in
      if sp_incl &&
      p_first_name base p = "?" && p_surname base p = "?" then
        do print_title ();
           html_li conf;
           Num.print wpr (transl conf "(thousand separator)") n;
           Wserver.wprint " -\n";
           if Array.length u.family > 0 then
             let cpl = coi base u.family.(0) in
             let (parent_name_index, conj) =
               match p.sex with
               [ Male -> (0, cpl.mother)
               | _ -> (1, cpl.father) ]
             in
             Wserver.wprint "%s"
               (transl_decline2 conf "%1 of %2"
                  (geneweb_link conf (acces conf base p)
                     (transl_nth conf "husband/wife" parent_name_index))
                  (person_title_text conf base (poi base conj) ^
                   Date.short_dates_text conf base (poi base conj)))
           else
             do afficher_personne_titre_referencee conf base p;
                Date.afficher_dates_courtes conf base p;
             return ();
           Wserver.wprint "\n";
        return ()
      else ()
  | GP_missing n ip ->
      let p = poi base ip in
      if p_first_name base p = "?" && p_surname base p = "?" then ()
      else
        let n1 = Num.twice n in
        let n2 = Num.inc n1 1 in
        do print_title ();
           html_li conf;
           Num.print wpr (transl conf "(thousand separator)") n1;
           Wserver.wprint "-";
           Wserver.wprint "%d" (Num.modl n2 10);
           Wserver.wprint " -\n";
           let s =
             referenced_person_title_text conf base p ^
             Date.short_dates_text conf base p
           in
           Wserver.wprint "%s\n"
             (if sp_incl then
                transl_decline2 conf "%1 of %2"
                  (nominative (transl conf "parents")) s
              else s);
        return ()
  | _ -> () ]
;

value one_year base p =
  match Adef.od_of_codate p.birth with
  [ Some (Dgreg d _) -> Some d.year
  | _ ->
      match Adef.od_of_codate p.baptism with
      [ Some (Dgreg d _) -> Some (annee d)
      | _ ->
          match date_of_death p.death with
          [ Some (Dgreg d _) -> Some (annee d)
          | _ ->
              match p.burial with
              [ Buried cod ->
                  match Adef.od_of_codate cod with
                  [ Some (Dgreg d _) -> Some (annee d)
                  | _ -> None ]
              | Cremated cod ->
                  match Adef.od_of_codate cod with
                  [ Some (Dgreg d _) -> Some (annee d)
                  | _ -> None ]
              | UnknownBurial -> None ] ] ] ]
;

value one_year_gp base =
  fun
  [ GP_person _ ip _ -> one_year base (poi base ip)
  | GP_same _ _ ip -> one_year base (poi base ip)
  | GP_interv _ -> None
  | GP_missing _ ip -> one_year base (poi base ip) ]
;

value print_missing_ancestors conf base v spouses_included p =
  let after = p_getint conf.env "after" in
  let before = p_getint conf.env "before" in
  let mark = Array.create (base.data.persons.len) Num.zero in
  let rec generation niveau gpl =
    if niveau > v + 1 then ()
    else if gpl <> [] then
      let title = ref (Some niveau) in
      let gpl_to_print =
        List.fold_left
          (fun gpl gp ->
             match (after, before) with
             [ (Some a1, Some a2) ->
                 match one_year_gp base gp with
                 [ Some a ->
                    if a >= a1 && a <= a2 then [gp :: gpl]
                    else gpl
                 | None -> gpl ]
             | (Some a1, None) ->
                 match one_year_gp base gp with
                 [ Some a -> if a >= a1 then [gp :: gpl] else gpl
                 | None -> gpl ]
             | (None, Some a2) ->
                 match one_year_gp base gp with
                 [ Some a -> if a <= a2 then [gp :: gpl] else gpl
                 | None -> gpl ]
             | (None, None) -> [gp :: gpl] ])
          [] gpl
      in
      let gpl_to_print = List.rev gpl_to_print in
      do List.iter
           (print_generation_missing_persons conf base title spouses_included)
           gpl_to_print;
         if title.val = None then Wserver.wprint "</ul>\n" else ();
      return
      let gpl = next_generation base mark gpl in
      generation (niveau + 1) gpl
    else ()
  in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "missing ancestors")
            (txt_fun raw_access conf base p)))
  in
  do header conf title;
     Wserver.wprint "%s" (capitale (text_to conf v));
     match after with
     [ Some a -> Wserver.wprint " %s %d" (transl conf "after") a
     | None -> () ];
     match before with
     [ Some a -> Wserver.wprint " %s %d" (transl conf "before") a
     | None -> () ];
     Wserver.wprint ".\n";
     if not spouses_included then
       do html_br conf; html_br conf; return
       Wserver.wprint "%s\n"
         (capitale
            (transl_decline2 conf "%1 of %2"
               (nominative (transl conf "parents")) "..."))
     else ();  
     mark.(Adef.int_of_iper p.cle_index) := Num.one;
     tag "ul" begin
       generation 1 [GP_person Num.one p.cle_index None];
     end;
     trailer conf;
  return ()
;

type missing_type =
  [ A_person
  | A_surname_of_husband_of of string
  | A_surname_of_wife_of of string
  | A_husband_of
  | A_wife_of
  | A_parents_of ]
;

value add_missing conf base spouses_included list =
  fun
  [ GP_person n ip _ ->
      let p = poi base ip in
      let u = uoi base ip in
      if spouses_included && p_first_name base p = "?"
      && p_surname base p = "?" then
        if Array.length u.family > 0 then
          let cpl = coi base u.family.(0) in
          let (a, p) =
            match p.sex with
            [ Male -> (A_husband_of, poi base cpl.mother)
            | _ -> (A_wife_of, poi base cpl.father) ]
          in
          [(a, p) :: list]
        else [(A_person, p) :: list]
      else list
  | GP_missing n ip ->
      let p = poi base ip in
      let u = uoi base ip in
      if spouses_included
      && (p_surname base p = "?" || p_surname base p = "N..." ) then
        if p_first_name base p = "?" then list
        else
          if Array.length u.family > 0 then
            let n = person_text_without_surname conf base p in
            let cpl = coi base u.family.(0) in
            let (a, p) =
              match p.sex with
              [ Male -> (A_surname_of_husband_of n, poi base cpl.mother)
              | _ -> (A_surname_of_wife_of n, poi base cpl.father) ]
            in
            if p_surname base p = "?" then list else [(a, p) :: list]
          else [(A_parents_of, p) :: list]
      else if p_surname base p = "?" || p_surname base p = "?" then
        list
      else [(A_parents_of, p) :: list]
  | _ -> list ]
;

value val_of_mt =
  fun
  [ A_person -> 0
  | A_surname_of_husband_of _ -> 1
  | A_surname_of_wife_of _ -> 2
  | A_husband_of -> 3
  | A_wife_of -> 4
  | A_parents_of -> 5 ]
;

value compare base (mt1, p1) (mt2, p2) =
  let c = alphabetique (p_surname base p1) (p_surname base p2) in
  if c == 0 then
    let c =
      alphabetique (p_first_name base p1) (p_first_name base p2)
    in
    if c == 0 then
      if p1 == p2 then val_of_mt mt1 < val_of_mt mt2
      else
        match (Adef.od_of_codate p1.birth, Adef.od_of_codate p2.birth) with
        [ (Some d1, Some d2) -> d1 strictement_avant d2
        | _ -> p1.occ < p2.occ ]
    else c < 0
  else c > 0
;

value print_missing_type conf =
  fun
  [ A_person -> ()
  | A_surname_of_husband_of x ->
      Wserver.wprint "%s %s"
        (transl_decline2 conf "%1 of %2" (transl_nth conf "surname/surnames" 0)
           (transl_nth conf "husband/wife" 0))
        x
  | A_surname_of_wife_of x ->
      Wserver.wprint "%s %s"
        (transl_decline2 conf "%1 of %2" (transl_nth conf "surname/surnames" 0)
           (transl_nth conf "husband/wife" 1))
        x
  | A_husband_of ->
      Wserver.wprint "%s" (transl_nth conf "husband/wife" 0) 
  | A_wife_of ->
      Wserver.wprint "%s" (transl_nth conf "husband/wife" 1)
  | A_parents_of ->
      Wserver.wprint "%s" (nominative (transl conf "parents")) ]
;

value print_spouses conf base p u =
  Array.iter
    (fun ifam ->
       let fam = foi base ifam in
       let cpl = coi base ifam in
       let sp = poi base (spouse p.cle_index cpl) in
       if p_first_name base sp = "?" && p_surname base sp = "?" then ()
       else
         do Wserver.wprint "\n&amp;%s\n"
              (Date.short_marriage_date_text conf base fam p sp);
            afficher_personne_titre conf base sp;
            Date.afficher_dates_courtes conf base sp;
         return ())
    u.family
;

value print_someone_missing conf base begin_surname spouses_incl (mt, mtl, p) =
  do let href= "i=" ^ string_of_int (Adef.int_of_iper p.cle_index) in
     wprint_geneweb_link conf href (person_text_without_surname conf base p);
     Wserver.wprint "%s" begin_surname;
     afficher_titre conf base p;
     Date.afficher_dates_courtes conf base p;
     if spouses_incl then
       do Wserver.wprint "\n=&gt; ";
          print_missing_type conf mt;
          List.iter
            (fun mt ->
               do Wserver.wprint ", ";
                  print_missing_type conf mt;
               return ())
            mtl;
       return ()
     else
       print_spouses conf base p (uoi base p.cle_index);
  return ()
;

value print_alphabetic_missing conf base spouses_included (surname, list) =
  do Wserver.wprint "%s " (surname_end surname);
     match list with
     [ [e] ->
         print_someone_missing conf base (surname_begin surname)
            spouses_included e
     | _ ->
        do Wserver.wprint "%s\n" (surname_begin surname);
           tag "ul" begin
             List.iter
               (fun e ->
                  do html_li conf;
                     print_someone_missing conf base "" spouses_included e;
                     Wserver.wprint "\n";
                  return ())
               list;
           end;
        return () ];
  return ()
;

value print_missing_ancestors_alphabetically conf base v spouses_included p =
  let mark = Array.create (base.data.persons.len) Num.zero in
  let rec generation list niveau gpl =
    if niveau > v then list
    else if gpl <> [] then
      let list =
        List.fold_left (add_missing conf base spouses_included) list gpl
      in
      let gpl = next_generation base mark gpl in
      generation list (niveau + 1) gpl
    else list
  in
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "missing ancestors")
         (txt_fun raw_access conf base p)))
  in
  let after = p_getint conf.env "after" in
  let before = p_getint conf.env "before" in
  do header conf title;
     let list = generation [] 1 [GP_person Num.one p.cle_index None] in
     let list =
       List.fold_left
         (fun npl (n, p) ->
            match (after, before) with
            [ (Some a1, Some a2) ->
                match one_year base p with
                [ Some a ->
                   if a >= a1 && a <= a2 then [(n, p) :: npl]
                   else npl
                | None -> npl ]
            | (Some a1, None) ->
                match one_year base p with
                [ Some a -> if a >= a1 then [(n, p) :: npl] else npl
                | None -> npl ]
            | (None, Some a2) ->
                match one_year base p with
                [ Some a -> if a <= a2 then [(n, p) :: npl] else npl
                | None -> npl ]
            | (None, None) -> [(n, p) :: npl] ])
         [] list
     in
     let list = Sort.list (compare base) list in
     let list =
       List.fold_left
         (fun nell ((_, p) as elm) ->
            match nell with
            [ [(n, el) :: nell] when n = p_surname base p ->
                [(n, [elm :: el]) :: nell]
            | _ -> [(p_surname base p, [elm]) :: nell] ])
         [] list
     in
     let list =
       List.map
         (fun (n, el) ->
            let ell =
              List.fold_left
                (fun ell ((a, p) as e) ->
                   match ell with
                   [ [(a1, al, p1) :: el] when p1 == p ->
                       [(a, [a1:: al], p) :: el]
                   | _ -> [(a, [], p) :: ell] ])
                [] el
            in
            (n, ell))
         list
     in
     let initials =
       List.fold_left
         (fun l (n, _) ->
            let i = n.[initiale n] in
            match l with
            [ [] -> [i]
            | [x :: l'] -> if x = i then l else [i :: l] ])
         [] list
     in
     let print_initials =
       List.length initials > 3 && List.length list > 100
     in
     do if print_initials then
          do html_p conf;
             List.iter
                (fun i ->
                   do stag "a" "href=\"#%c\"" i begin
                        Wserver.wprint "%c" i;
                      end;
                      Wserver.wprint "\n";
                   return ())
                (List.rev initials);
             html_p conf;
          return ()
        else ();
        Wserver.wprint "%s" (capitale (text_to conf v));
        match after with
        [ Some a -> Wserver.wprint " %s %d" (transl conf "after") a
        | None -> () ];
        match before with
        [ Some a -> Wserver.wprint " %s %d" (transl conf "before") a
        | None -> () ];
        Wserver.wprint ".\n";
        if not spouses_included then
          do html_br conf; html_br conf; return
          Wserver.wprint "%s\n"
            (capitale
               (transl_decline2 conf "%1 of %2"
                   (nominative (transl conf "parents")) "..."))
        else ();  
        tag "ul" begin
          let _ = List.fold_left
            (fun prev_i ((n, _) as e) ->
               let i = n.[initiale n] in
               do if print_initials then
                    match prev_i with
                    [ Some pi ->
                        if i <> pi then
                          do Wserver.wprint "</ul>\n";
                             html_li conf;
                             Wserver.wprint "<a name=\"%c\">%c</a>\n" i i;
                             Wserver.wprint "<ul>\n";
                          return ()
                        else ()
                    | None ->
                        do html_li conf;
                           Wserver.wprint "<a name=\"%c\">%c</a>\n" i i;
                           Wserver.wprint "<ul>\n";
                        return () ]
                  else ();
                  html_li conf;
                  print_alphabetic_missing conf base spouses_included e;
                  Wserver.wprint "\n";
               return Some i)
            None list
          in ();
          if print_initials then Wserver.wprint "</ul>\n" else ();
        end;
     return ();
     trailer conf;
  return ()
;

value tree_reference gv conf base p s =
  if conf.cancel_links then s
  else
    let im = p_getenv conf.env "image" = Some "on" in
    "<a href=\"" ^ commd conf ^ "m=A;t=T;v=" ^ string_of_int gv ^ ";" ^
    acces conf base p ^ (if im then ";image=on" else "") ^ "\">" ^ s ^ "</a>"
;

(*
value someone_text conf base reference p =
  reference conf base p (person_title_text conf base p) ^
  Date.short_dates_text conf base p
;

value print_centered_on sz pos s =
  let len = pre_text_size s in
  let sh =
    if len / 2 < pos then
      if pos + len / 2 < sz then pos - len / 2
      else sz - 1 - len
    else 1
  in
  Wserver.wprint "%s%s\n" (String.make sh ' ') s
;

value print_tree_with_pre conf base v p =
  let sz = 79 in
  let v = min 3 v in
  let print_parents sz pf pm ifath imoth =
    let fath = poi base ifath in
    let moth = poi base imoth in
    do print_centered_on sz pm
         (someone_text conf base (tree_reference v) moth);
       print_centered_on sz pf
         (someone_text conf base (tree_reference v) fath);
    return ()
  in
  let str2 = String.make (sz / 2) in
  let str4 = String.make (sz / 4) in
  tag "pre" begin
     match (aoi base p.cle_index).parents with
     [ Some ifam ->
         let cpl = coi base ifam in
          do if v >= 3 then
              do match (aoi base cpl.mother).parents with
                 [ Some ifam ->
                     let cpl = coi base ifam in
                     print_parents sz (5 * sz / 8) (7 * sz / 8)
                       cpl.father cpl.mother
                 | None -> () ];
                 match (aoi base cpl.father).parents with
                 [ Some ifam ->
                     let cpl = coi base ifam in
                     print_parents sz (sz / 8) (3 * sz / 8)
                       cpl.father cpl.mother
                 | None -> () ];
                 print_pre_center sz 
                   ("|" ^ str4 ' ' ^ "|" ^ str4 ' ' ^ "|" ^ str4 ' ' ^ "|");
                 print_pre_center sz
                   (str4 '_' ^ " " ^ str4 ' ' ^ " " ^ str4 '_');
                 print_pre_center sz ("|" ^ str2 ' ' ^ "|");
              return ()
            else ();
            print_parents sz (sz / 4) (3 * sz / 4) cpl.father cpl.mother;
         return ()
     | None -> () ];
     print_pre_center sz ("|" ^ str2 ' ' ^ "|");
     print_pre_center sz (str2 '_');
     print_pre_center sz "|";
     print_pre_center sz (someone_text conf base reference p);
  end
;
*)

type pos = [ Left | Right | Center | Alone ];

type cell = [ Cell of person and
                      pos and  (* Position for horizontal bar *)
                      bool and (* Top vertical bar needed ? *)
                      int      (* Colspan *)
	    | Empty ];

(* Ascendant tree:

  8 ? ? ? ? ? ? ?
   4   5   ?   7
     2       3 
         1

1) Build list of levels (t1 = True for parents flag, size 1)
   => [ [8At1 E E] [4Lt1 5Rt1 7At1] [2Lt1 3Rt1] [1Ct1] ] 

2) Enrich list of levels (parents flag, sizing)
   => [ [8At1 E E] [4Lt1 5Rf1 7Af1] [2Lt3 3Rt1] [1Ct5] ]

3) Display it
    For each cell:
      Top vertical bar if parents flag (not on top line)
      Person
      Person tree link (vertical bar) ) not on bottom line
      Horizontal line                 )

*)

value rec enrich lst1 lst2 =
  match (lst1, lst2) with
  [ (_, []) -> []
  | ([], lst) -> lst
  | (_, _) ->
      match (List.hd lst1, List.hd lst2) with
      [ (Cell _ Right _ s1, Cell p d u s2) ->
          [(Cell p d u (s1+s2+1)) :: (enrich (List.tl lst1) (List.tl lst2))]
      | (Cell _ Left _ s, Cell p d u _) ->
          enrich (List.tl lst1) [Cell p d u s :: List.tl lst2]
      | (Cell _ _ _ s, Cell p d u _) ->
          [Cell p d u s :: enrich (List.tl lst1) (List.tl lst2)]
      | (Empty, Cell p d _ s) ->
          [ Cell p d False s :: enrich (List.tl lst1) (List.tl lst2)]
      | (_, Empty) ->
          [ Empty :: enrich (List.tl lst1) (List.tl lst2)] ]
  ]
;
    
value is_empty lst =
  List.fold_left
    (fun test po -> test && (po == Empty)
    ) True lst
;
     
value rec enrich_tree lst =
  match lst with
  [ [] -> []
  | [head :: tail] ->
      if is_empty head then
        enrich_tree tail
      else
        match tail with
        [ [] -> [head]
        | [thead :: ttail] ->
            [head :: enrich_tree [(enrich head thead) :: ttail]] ]
  ]
;

(* print_tree_with_table:
    conf: configuration parameters
    base: base name
    gv: number of generations
    p: person *)
value print_tree_with_table conf base gv p =
  let gv = min (limit_by_tree conf) gv in
  let next_gen pol =
    List.fold_right
      (fun po list ->
         match po with
         [ Empty -> [Empty :: list]
         | Cell p _ _ _ ->
             match (aoi base p.cle_index).parents with
             [ Some ifam ->
                 let cpl = coi base ifam in
                 let fath =
                   let p = poi base cpl.father in
                   if connais base p then Some p else None
                 in
                 let moth =
                   let p = poi base cpl.mother in
                   if connais base p then Some p else None
                 in
		 match (fath, moth) with
		 [ (Some f, Some m) ->
                     [Cell f Left True 1; Cell m Right True 1 :: list]
		 | (Some f, None)   -> [Cell f Alone True 1 :: list]
                 | (None, Some m)   -> [Cell m Alone True 1 :: list]
		 | (None, None)     -> [Empty :: list] ]
             | _ -> [Empty :: list] ] ])
      pol []
  in
  (* build tree: go up (initialisation) *)
  let gen =
    loop (gv - 1) [Cell p Center True 1] [] where rec loop i gen list =
      if i == 0 then [gen :: list]
      else loop (i - 1) (next_gen gen) [gen :: list]
  in
  (* build tree: go down (enrichment) *)
  let gen = enrich_tree gen in
  let down_reference p s =
    if conf.cancel_links then s
    else reference conf base p s
  in
  let colspan = 
    fun [ Empty 
        | Cell _ _ _ 1 -> ""
        | Cell _ _ _ s -> " colspan=" ^ string_of_int s ] in
  let align =
    fun [ Cell _ Center _ _
        | Cell _ Alone _ _  -> "align=center"
        | Cell _ Left _ _   -> "align=right"
        | _                 -> "" ]
  in
  let print_ancestor_link gen first po =
    do if not first then Wserver.wprint "<td>&nbsp;&nbsp;</td>\n"
       else ();
       stag "td" "align=center%s" (colspan po) begin
         let txt =
           match po with
           [ Empty -> "&nbsp;"
           | Cell p _ _ _ ->
               tree_reference gv conf base p "|" ]
         in
         Wserver.wprint "%s" txt;
       end;
       Wserver.wprint "\n";
    return ()
  in
  let print_ancestor gen first po =
    do if not first then Wserver.wprint "<td>&nbsp;&nbsp;</td>\n"
       else ();
       stag "td" "align=center%s" (colspan po) begin
         let txt =
           match po with
           [ Empty -> "&nbsp;"
           | Cell p _ _ _ ->
               let txt = person_title_text conf base p in
               let txt = down_reference p txt in
                   txt ^ Date.short_dates_text conf base p ]
         in
         Wserver.wprint "%s" txt;
         match po with
         [ Empty -> ()
         | Cell p _ _ _ -> Wserver.wprint "%s" (Dag.image_txt conf base p) ];
       end;
       Wserver.wprint "\n";
    return ()
  in
  let print_vertical_bars gen first po =
    do if not first then Wserver.wprint "<td>&nbsp;&nbsp;</td>\n"
       else ();
       stag "td" "align=center%s" (colspan po) begin
         let txt =
           match po with
           [ Empty 
           | Cell _ _ False _ -> "&nbsp;"
           | _ -> "|" ]
         in
         Wserver.wprint "%s" txt;
       end;
       Wserver.wprint "\n";
    return ()
  in
  let print_horizontal_line gen first po =
    do if not first then
         do stag "td" "%s" (align po) begin
              let txt = 
                match po with
                [ Cell _ Right _ _
                | Cell _ Center _ _ -> "<hr noshade size=1>"
                | _ -> "&nbsp;" ]
              in
              Wserver.wprint "%s" txt;
            end;
            Wserver.wprint "\n";
          return ()
       else ();
       stag "td" "%s%s" (align po) (colspan po) begin
         let txt =
           match po with
           [ Empty -> "&nbsp;"
           | Cell _ Left _ _ ->
              "<hr noshade size=1 width=\"50%%\" align=right>"
           | Cell _ Right _ _ ->
              "<hr noshade size=1 width=\"50%%\" align=left>"
           | Cell _ Alone _ _ ->
              "|"
           | Cell _ Center _ _ ->
              "<hr noshade size=1>" ]
         in
         Wserver.wprint "%s" txt;
       end;
       Wserver.wprint "\n";
    return ()
  in
  tag "table" "border=%d cellspacing=0 cellpadding=0 width=\"100%%\""
    conf.border
  begin
    list_iter_first
      (fun firstline gen ->
         do if not firstline then
              tag "tr" begin
                list_iter_first
                  (fun first po -> print_vertical_bars gen first po) gen;
              end
            else ();
            tag "tr" begin
              list_iter_first (fun first po -> print_ancestor gen first po)
                gen;
            end;
            match gen with
            [ [ Cell _ Center _ _ :: _ ] -> ()
            | _ ->
              do tag "tr" begin
                   list_iter_first
                     (fun first po -> print_ancestor_link gen first po) gen;
                 end;
                 tag "tr" begin
                   list_iter_first
                      (fun first po -> print_horizontal_line gen first po) gen;
                 end;
              return ()
            ];
         return ())
      gen;
  end
;

value print_normal_tree conf base v p =
  let title _ =
    Wserver.wprint "%s: %s" (capitale (transl conf "tree"))
      (person_text_no_html conf base p)
  in
  do header_no_page_title conf title;
     Wserver.wprint "<div align=right><a href=\"%s" (commd conf);
     List.iter (fun (k, v) -> Wserver.wprint "%s=%s;" k v) conf.env;
     Wserver.wprint "dag=on;notab=on;slices=on";
     Wserver.wprint "\"><tt>//</tt></a></div>\n";
     print_tree_with_table conf base v p;
     trailer conf;
  return ()
;

value print_tree conf base v p =
  let v = min (limit_by_tree conf) v in
  if p_getenv conf.env "dag" = Some "on"
  || browser_doesnt_have_tables conf then
    let set =
      loop Dag.Pset.empty v p.cle_index where rec loop set lev ip =
        let set = Dag.Pset.add ip set in
        if lev <= 1 then set
        else
          match (aoi base ip).parents with
          [ Some ifam ->
              let cpl = coi base ifam in
              let set = loop set (lev - 1) cpl.mother in
              loop set (lev - 1) cpl.father
          | None -> set ]
    in
    let d = Dag.make_dag conf base (Dag.Pset.elements set) in
    Dag.gen_print_dag conf base False True set [] d
  else
    print_normal_tree conf base v p
;

value no_spaces s =
  loop 0 0 where rec loop len i =
    if i == String.length s then Buff.get len
    else
      let len =
        match s.[i] with
        [ ' ' -> Buff.mstore len "&nbsp;"
        | x -> Buff.store len s.[i] ]
      in
      loop len (i + 1)
;

value htree_reference gv conf base p s =
  if conf.cancel_links then s
  else
    "<a href=\"" ^ commd conf ^ "m=A;t=H;v=" ^ string_of_int gv ^ ";" ^
    acces conf base p ^ "\">" ^ s ^ "</a>"
;

value print_horizontally conf base max_level p =
  let title h =
    let txt_fun = if h then gen_person_text_no_html else gen_person_text in
    Wserver.wprint "%s"
      (capitale
         (transl_decline2 conf "%1 of %2" (transl conf "ancestors")
            (txt_fun raw_access conf base p)))
  in
  let max_level = min (limit_by_list conf) max_level in
  let suff1 = "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" in
  let suff2 = "&nbsp;+--&nbsp;" in
  let suff3 = "&nbsp;|&nbsp;&nbsp;&nbsp;" in
  let rec loop level s1 s2 s3 ip =
    if level >= max_level then ()
    else
      do match (aoi base ip).parents with
         [ Some ifam ->
             loop (level + 1) (s1 ^ suff1) (s1 ^ suff2) (s1 ^ suff3)
               (coi base ifam).father
         | None -> () ];
         Wserver.wprint "<tt>%s</tt>" s2;
         let p = poi base ip in
         let ref =
           if level = 0 then reference else htree_reference max_level
         in
         Wserver.wprint "%s%s<br>\n"
           (ref conf base p (no_spaces (person_text conf base p)))
           (no_spaces (Date.short_dates_text conf base p));
         match (aoi base ip).parents with
         [ Some ifam ->
             loop (level + 1) (s3 ^ suff3) (s3 ^ suff2) (s3 ^ suff1)
               (coi base ifam).mother
         | None -> () ];
      return ()
  in
  do header conf title;
     print_link_to_welcome conf True;
     Wserver.wprint "%s.\n" (capitale (text_to conf max_level));
     Wserver.wprint "<p>\n";
     tag "nobr" begin
       let suff13 = "&nbsp;&nbsp;&nbsp;" in
       let suff2 = "--&nbsp;" in
       loop 0 suff13 suff2 suff13 p.cle_index;
     end;
     trailer conf;
  return ()
;

value print_male_female_line male conf base v p =
  let list =
    loop [] v p.cle_index where rec loop list lev ip =
      let list = [ip :: list] in
      if lev <= 1 then list
      else
        match (aoi base ip).parents with
        [ Some ifam ->
            let cpl = coi base ifam in
            loop list (lev - 1) (if male then cpl.father else cpl.mother)
        | None -> list ]
  in
  let title _ =
    Wserver.wprint "%s: %s"
      (capitale
         (transl_nth conf "male line/female line" (if male then 0 else 1)))
      (person_text_no_html conf base p)
  in
  do header_no_page_title conf title;
     tag "center" begin
       list_iter_first
         (fun first ip ->
            let p = poi base ip in
            do if not first then Wserver.wprint "|<br>\n" else ();
               Wserver.wprint "%s\n%s<br>\n"
                 (referenced_person_title_text conf base p)
                 (Date.short_dates_text conf base p);
               Wserver.wprint "%s" (Dag.image_txt conf base p);
            return ())
         list;
     end;
     trailer conf;
  return ()
;

value print_male_line = print_male_female_line True;
value print_female_line = print_male_female_line False;

(* Surnames list *)

value get_date_place conf base auth_for_all_anc p =
  if auth_for_all_anc || age_autorise conf base p then
    let d1 =
      match Adef.od_of_codate p.birth with
      [ None -> Adef.od_of_codate p.baptism
      | x -> x ]
    in
    let d1 =
      if d1 <> None then d1
      else
        List.fold_left
          (fun d ifam ->
             if d <> None then d
             else Adef.od_of_codate (foi base ifam).marriage)
          d1 (Array.to_list (uoi base p.cle_index).family)
      
    in
    let d2 =
      match p.death with
      [ Death _ cd -> Some (Adef.date_of_cdate cd)
      | _ ->
          match p.burial with
          [ Buried cod -> Adef.od_of_codate cod
          | Cremated cod -> Adef.od_of_codate cod
          | _ -> None ] ]
    in
    let auth_for_all_anc =
      if auth_for_all_anc then True
      else
        match d2 with
        [ Some (Dgreg d _)
          when (temps_ecoule d conf.today).year > conf.private_years -> True
        | _ -> False ]
    in
    let pl =
      let pl = "" in
      let pl = if pl <> "" then pl else sou base p.birth_place in
      let pl = if pl <> "" then pl else sou base p.baptism_place in
      let pl = if pl <> "" then pl else sou base p.death_place in
      let pl = if pl <> "" then pl else sou base p.burial_place in
      let pl =
        if pl <> "" then pl
        else
          List.fold_left
            (fun pl ifam ->
               if pl <> "" then pl
               else sou base (foi base ifam).marriage_place)
            pl (Array.to_list (uoi base p.cle_index).family)
      in
      pl
    in
    ((d1, d2, pl), auth_for_all_anc)
  else ((None, None, ""), False)
;

value merge_date_place conf base surn ((d1, d2, pl), auth) p =
  let ((pd1, pd2, ppl), auth) = get_date_place conf base auth p in
  let nd1 =
    if pd1 <> None then pd1
    else if p.surname = surn then
      if pd2 <> None then pd2 else d1
    else None
  in
  let nd2 =
    if p.surname = surn then
      if d2 <> None then d2
      else if d1 <> None then d1
      else if pd1 <> None then pd2 else pd1
    else
      if pd2 <> None then pd2
      else if pd1 <> None then pd1
      else d1
  in
  let pl =
    if ppl <> "" then ppl
    else if p.surname = surn then pl
    else ""
  in
  ((nd1, nd2, pl), auth)
;

value build_surnames_list conf base v p =
  let ht = Hashtbl.create 701 in
  let mark = Array.create base.data.persons.len 5 in
  let auth = conf.wizard || conf.friend in
  let add_surname sosa p surn dp =
    let r =
      try Hashtbl.find ht surn with
      [ Not_found ->
          let r = ref ((fst dp, p), []) in
          do Hashtbl.add ht surn r; return r ]
    in
    r.val := (fst r.val, [sosa :: snd r.val])
  in
  do loop 0 (Num.one) p p.surname (get_date_place conf base auth p)
     where rec loop lev sosa p surn dp =
       if mark.(Adef.int_of_iper p.cle_index) = 0 then ()
       else if lev = v then add_surname sosa p surn dp
       else
         do mark.(Adef.int_of_iper p.cle_index) :=
              mark.(Adef.int_of_iper p.cle_index) - 1;
            match (aoi base p.cle_index).parents with
            [ Some ifam ->
                let cpl = coi base ifam in
                let fath = poi base cpl.father in
                let moth = poi base cpl.mother in
                do if surn <> fath.surname && surn <> moth.surname then
                     add_surname sosa p surn dp
                   else ();
                return
                let sosa = Num.twice sosa in
                let dp1 = merge_date_place conf base surn dp fath in
                do loop (lev + 1) sosa fath fath.surname dp1; return
                let sosa = Num.inc sosa 1 in
                let dp2 = merge_date_place conf base surn dp moth in
                do loop (lev + 1) sosa moth moth.surname dp2; return ()
            | None ->
                add_surname sosa p surn dp ];
         return ();
  return
  let list = ref [] in
  do Hashtbl.iter
       (fun i dp ->
          let surn = sou base i in
          if surn <> "?" then list.val := [(surn, dp.val) :: list.val]
          else ())
       ht;
  return
  Sort.list (fun (s1, _) (s2, _) -> Gutil.alphabetique s1 s2 <= 0) list.val
;

value print_surnames_list conf base v p =
  let title h =
     do if not h then Wserver.wprint "%s<br>" (person_text conf base p)
        else ();
        Wserver.wprint "- %s -" (capitale (transl conf "surnames list"));
     return ()
  in
  let list = build_surnames_list conf base v p in
  let with_tab =
    not (Util.browser_doesnt_have_tables conf) &&
    p_getenv conf.env "tab" = Some "on"
  in
  do Util.header conf title;
     Util.print_link_to_welcome conf True;
     if with_tab then Wserver.wprint "<br>\n" else ();
     Wserver.wprint "<%s>\n"
       (if with_tab then "table border=1 width=90%%" else "ul");
     List.iter
       (fun (surn, (((d1, d2, pl), anc), sosa_list)) ->
          let d2 = if d2 = d1 then None else d2 in
          do Wserver.wprint "<%s>\n" (if with_tab then "tr><td" else "li");
             if Util.browser_doesnt_have_tables conf then
               let sosa = List.hd (List.rev sosa_list) in
               wprint_geneweb_link conf
                 ("m=RL;" ^ acces conf base anc ^ ";" ^
                  acces_n conf base "1" p ^ ";" ^
                  acces_n conf base "2" anc ^ ";b1=" ^
                  Num.to_string sosa ^ ";b2=1")
                 (surname_end surn ^ surname_begin surn)
             else
               let (str, _) =
                 List.fold_right
                   (fun sosa (str, n) ->
                      let str =
                        str ^ ";s" ^ string_of_int n ^ "=" ^
                        Num.to_string sosa
                      in
                      (str, n + 1))
                   sosa_list ("", 1)
               in
               wprint_geneweb_link conf
                 ("m=DAG;" ^ acces_n conf base "1" p ^ str)
                 (surname_end surn ^ surname_begin surn);
             if conf.cancel_links then ()
             else
               let comm =
                 match List.length sosa_list with
                 [ 1 -> ""
                 | n -> " (" ^ string_of_int n ^ ")" ]
               in
               Wserver.wprint "%s" comm;
             Wserver.wprint "%s&nbsp;%s%s" (if with_tab then "<td>" else ";")
               pl (if with_tab then "<td>" else ";");
             Wserver.wprint "&nbsp;";
             match d1 with
             [ Some (Dgreg d _) -> Wserver.wprint "%d" d.year
             | Some (Dtext s) -> Wserver.wprint "%s" s
             | None -> () ];
             if d1 <> None && d2 <> None then Wserver.wprint "-" else ();
             match d2 with
             [ Some (Dgreg d _) -> Wserver.wprint "%d" d.year
             | Some (Dtext s) -> Wserver.wprint "%s" s
             | None -> () ];
             Wserver.wprint "\n";
          return ())
       list;
     Wserver.wprint "</%s>\n" (if with_tab then "table" else "ul");
     Util.trailer conf;
  return ()
;

value print conf base p =
  match (p_getenv conf.env "t", p_getint conf.env "v") with
  [ (Some "L", Some v) -> afficher_ascendants_jusqu_a conf base v p
  | (Some "N", Some v) -> afficher_ascendants_numerotation conf base v p
  | (Some "G", Some v) ->
      let ws =
        match p_getenv conf.env "siblings" with
        [ Some "on" -> True
        | _ -> False ]
      in
      let wn =
        match p_getenv conf.env "notes" with
        [ Some "on" -> True
        | _ -> False ]
      in
      afficher_ascendants_numerotation_long conf base v ws wn p
  | (Some "S", Some v) -> afficher_ascendants_niveau conf base v p
  | (Some "M", Some v) ->
      let al =
        match p_getenv conf.env "al" with
        [ Some "on" -> True
        | _ -> False ]
      in
      let si =
        match p_getenv conf.env "ms" with
        [ Some "on" -> True
        | _ -> False ]
      in
      if al then print_missing_ancestors_alphabetically conf base v si p
      else print_missing_ancestors conf base v si p
  | (Some "T", Some v) -> print_tree conf base v p
  | (Some "H", Some v) -> print_horizontally conf base v p
  | (Some "A", Some v) -> print_male_line conf base v p
  | (Some "C", Some v) -> print_female_line conf base v p
  | (Some "D", x) ->
      match (find_person_in_env conf base "1", x) with
      [ (Some anc, _) ->
          print_ancestors_same_time_descendants conf base p anc
      | (_, Some v) -> (* compatibility *)
          print_ancestors_same_time_descendants conf base p
            (base.data.persons.get v)
      | _ -> afficher_menu_ascendants conf base p ]
  | (Some "F", Some v) -> print_surnames_list conf base v p
  | _ -> afficher_menu_ascendants conf base p ]
;
